/***************************************************************************
                          s_internal_loop.cpp  -  description
                             -------------------
    begin                : Fri Apr 12 2002
    copyright            : (C) 2002 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// a class for internal loop related functions
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "s_internal_loop.h"
#include "externs.h"
#include "common.h"
#include "simfold.h"
#include "params.h"


s_internal_loop::s_internal_loop (int *seq, int length)
// The constructor
{
    seqlen = length;
    sequence = seq;
    this->V = NULL;
}


s_internal_loop::~s_internal_loop ()
// The destructor
{
}

int s_internal_loop::compute_energy (int i, int j)
// computes the MFE of the structure closed by an internal loop closed at (i,j)
{

    int ip, jp, minq;
    int mmin, ttmp;
    int penalty_size, asym_penalty, ip_jp_energy, i_j_energy, en;
    int branch1, branch2, l;
    mmin = INF;

    for (ip = i+1; ip <= MIN(j-2-TURN,i+MAXLOOP+1) ; ip++)  // j-2-TURN
    {
        minq = MAX (j-i+ip-MAXLOOP-2, ip+1+TURN);    // ip+1+TURN);
        for (jp = minq; jp < j; jp++)
        {        
            // could replace the code from now to the end of the function, by get_energy_str
            // it's very redundant (duplicated code) as it is now, but it's faster.
            if (sequence[ip]+sequence[jp] == 3 ||
                sequence[ip]+sequence[jp] == 5)        
            {
            
                branch1 = ip-i-1;
                branch2 = j-jp-1;
                if (branch1 == 0 && branch2 == 0)
                    continue;

                if (branch1 != 0 || branch2 != 0)
                {
                    // check if it is a bulge loop of size 1
                    // check if it is int11 or int21 or int22
                    // 1x1 internal loop
                    if (branch1 == 1 && branch2 == 1 && !simple_internal_energy)     // it is int11
                    {
                        // int11[i][j][i+1][j-1][ip][jp]
                        en = int11 [sequence[i]][sequence[j]]
                                   [sequence[i+1]][sequence[j-1]]
                                   [sequence[ip]][sequence[jp]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                        continue;
                    }
                    // 1x2 internal loop
                    else if (branch1 == 1 && branch2 == 2 && !simple_internal_energy)
                    {
                        // int21[i][j][i+1][j-1][ip][jp][jp+1]
                        en = int21 [sequence[i]][sequence[j]]
                                   [sequence[i+1]][sequence[j-1]]
                                   [sequence[ip]][sequence[jp]]
                                   [sequence[jp+1]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }

                        continue;
                    }
                    // 1x2 internal loop
                    else if(branch1 == 2 && branch2 == 1 && !simple_internal_energy)
                    {
                        // after rotation: int21[jp][ip][j-1][ip-1][j][i][i+1]
                        en = int21 [sequence[jp]][sequence[ip]]
                                   [sequence[j-1]][sequence[ip-1]]
                                   [sequence[j]][sequence[i]]
                                   [sequence[i+1]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                        continue;
                    }
                    // 2x2 internal loop
                    else if (branch1 == 2 && branch2 == 2 && !simple_internal_energy)
                    {
                        // int22[i][j][i+1][j-1][ip][jp][ip-1][jp+1]
                        en = int22 [sequence[i]][sequence[j]]
                                   [sequence[i+1]][sequence[j-1]]
                                   [sequence[ip]][sequence[jp]]
                                   [sequence[ip-1]][sequence[jp+1]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                        continue;
                    }
                    else
                    {
                        // this case is not int11, int21, int22
                        
                        // check if it is a bulge
                        if (branch1 == 0 || branch2 == 0)
                        {
                            l = branch1+branch2;
                            penalty_size = penalty_by_size (l, 'B');
                            if (l == 1)
                            {
                                // bulge of size 1
                                // stack[i][j][i+1][j-1]
                                en = stack [sequence[i]][sequence[j]]
                                           [sequence[ip]][sequence[jp]];
                                ttmp = en + penalty_size + V->get_energy (ip, jp);
                                if (ttmp < mmin)
                                {
                                    mmin = ttmp;
                                }
                                continue;
                            }
                            else
                            {
                                // bulge of size bigger than 1
                                // check if (i,j) and (ip,jp) can pair
                                ttmp = penalty_size +
                                    AU_penalty (sequence[i],sequence[j]) +
                                    AU_penalty (sequence[ip], sequence[jp]) +
                                    V->get_energy (ip, jp);
                                if (ttmp < mmin)
                                {
                                    mmin = ttmp;
                                }
                                continue;
                            }
                        }
                        // it is a regular internal loop (not a bulge)
                        else
                        {
                            l = branch1+branch2;
                            penalty_size = penalty_by_size (l, 'I');
                            asym_penalty = asymmetry_penalty (branch1, branch2);

                            if ((branch1 == 1 || branch2 == 1) && misc.gail_rule)
                            // If gail_rule is set to 1 in miscloop file,
                            // i_j_energy and ip_jp_energy will be calculated as if it was a loop of As
                            {
                                i_j_energy  =  tstacki[sequence[i]][sequence[j]]
                                                      [0][0];
                                ip_jp_energy = tstacki[sequence[jp]][sequence[ip]]
                                                      [0][0];
                            }
                            else
                            {
                                i_j_energy   = tstacki[sequence[i]][sequence[j]]
                                                      [sequence[i+1]][sequence[j-1]];
                                ip_jp_energy = tstacki[sequence[jp]][sequence[ip]]
                                                      [sequence[jp+1]][sequence[ip-1]];
                            }
                            ttmp = i_j_energy + ip_jp_energy + penalty_size +
                                         asym_penalty + V->get_energy (ip, jp);
                            if (ttmp < mmin)
                            {
                                mmin = ttmp;
                            }

                            continue;
                        }
                    }

                }
            }
        }
    }
    return mmin;
}


int s_internal_loop::compute_energy_restricted (int i, int j, str_features *fres)
// computes the MFE of the structure closed by a restricted internal loop closed by (i,j)
{

    int ip, jp, minq;
    int mmin, ttmp;
    int penalty_size, asym_penalty, ip_jp_energy, i_j_energy, en;
    int branch1, branch2, l;
    mmin = INF;

    for (ip = i+1; ip <= MIN(j-2,i+MAXLOOP+1) ; ip++)  // the -TURN shouldn't be there
    {
        minq = MAX (j-i+ip-MAXLOOP-2, ip+1);    // without TURN
        for (jp = minq; jp < j; jp++)
        {        
            if (exists_restricted (i,ip,fres) || exists_restricted (jp,j,fres))
                continue;
            ttmp = get_energy_str (i, j, ip, jp);
            if (ttmp < mmin)
            {
                mmin = ttmp;
            }
        }
    }
    return mmin;
}



int s_internal_loop::get_energy_str (int i, int j, int ip, int jp)
// returns the free energy of the structure closed by the internal loop (i,j,ip,jp)
// This function is just most of what is inside the second for loop of compute_energy
{
    int mmin, ttmp;
    int penalty_size, asym_penalty, ip_jp_energy, i_j_energy, en;
    int branch1, branch2, l;
    mmin = INF;
    
            if (sequence[ip]+sequence[jp] == 3 || sequence[ip]+sequence[jp] == 5)        
            {
            
                branch1 = ip-i-1;
                branch2 = j-jp-1;

                if (branch1 != 0 || branch2 != 0)
                {
                    // check if it is a bulge loop of size 1
                    // check if it is int11 or int21 or int22
                    if (branch1 == 1 && branch2 == 1 && !simple_internal_energy)     // it is int11
                    {
                        // int11[i][j][i+1][j-1][ip][jp]
                        en = int11 [sequence[i]][sequence[j]]
                                   [sequence[i+1]][sequence[j-1]]
                                   [sequence[ip]][sequence[jp]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                    }
                    else if (branch1 == 1 && branch2 == 2 && !simple_internal_energy)
                    {
                        // int21[i][j][i+1][j-1][ip][jp][jp+1]
                        en = int21 [sequence[i]][sequence[j]]
                                   [sequence[i+1]][sequence[j-1]]
                                   [sequence[ip]][sequence[jp]]
                                   [sequence[jp+1]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                    }
                    else if(branch1 == 2 && branch2 == 1 && !simple_internal_energy)
                    {
                        // after rotation: int21[jp][ip][j-1][ip-1][j][i][i+1]
                        en = int21 [sequence[jp]][sequence[ip]]
                                   [sequence[j-1]][sequence[ip-1]]
                                   [sequence[j]][sequence[i]]
                                   [sequence[i+1]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                    }
                    else if (branch1 == 2 && branch2 == 2 && !simple_internal_energy)
                    {
                        // int22[i][j][i+1][j-1][ip][jp][ip-1][jp+1]
                        en = int22 [sequence[i]][sequence[j]]
                                   [sequence[i+1]][sequence[j-1]]
                                   [sequence[ip]][sequence[jp]]
                                   [sequence[ip-1]][sequence[jp+1]];
                        ttmp = en + V->get_energy (ip, jp);
                        if (ttmp < mmin)
                        {
                            mmin = ttmp;
                        }
                    }
                    else
                    {
                        // this case is not int11, int21, int22
                        
                        // check if it is a bulge
                        if (branch1 == 0 || branch2 == 0)
                        {
                            l = branch1+branch2;
                            penalty_size = penalty_by_size (l, 'B');
                            if (l == 1)
                            {
                                // bulge of size 1
                                // stack[i][j][i+1][j-1]
                                en = stack [sequence[i]][sequence[j]]

                                           [sequence[ip]][sequence[jp]];
                                ttmp = en + penalty_size + V->get_energy (ip, jp);
                                if (ttmp < mmin)
                                {
                                    mmin = ttmp;
                                }
                            }
                            else
                            {
                                // bulge of size bigger than 1
                                // check if (i,j) and (ip,jp) can pair
                                ttmp = penalty_size +
                                    AU_penalty (sequence[i],sequence[j]) +
                                    AU_penalty (sequence[ip], sequence[jp]) +
                                    V->get_energy (ip, jp);
                                if (ttmp < mmin)
                                {
                                    mmin = ttmp;
                                }
                            }
                        }
                        // it is an internal loop (not a bulge)
                        else
                        {
                            l = branch1+branch2;
                            penalty_size = penalty_by_size (l, 'I');
                            asym_penalty = asymmetry_penalty (branch1, branch2);

                            if ((branch1 == 1 || branch2 == 1) && misc.gail_rule)
                            // If gail_rule is set to 1 in miscloop file,
                            // i_j_energy and ip_jp_energy will be calculated as if it was a loop of As
                            {
                                i_j_energy  =  tstacki[sequence[i]][sequence[j]]
                                                      [0][0];
                                ip_jp_energy = tstacki[sequence[jp]][sequence[ip]]
                                                      [0][0];
                            }
                            else
                            {
                                i_j_energy   = tstacki[sequence[i]][sequence[j]]
                                                      [sequence[i+1]][sequence[j-1]];
                                ip_jp_energy = tstacki[sequence[jp]][sequence[ip]]
                                                      [sequence[jp+1]][sequence[ip-1]];
                            }
                            ttmp = i_j_energy + ip_jp_energy + penalty_size +
                                         asym_penalty + V->get_energy (ip, jp);
                            if (ttmp < mmin)
                            {
                                mmin = ttmp;
                            }
                        }
                    }
                }
            }    
    if (mmin < INF/2)
        return mmin;
    return INF;            
}



int s_internal_loop::get_energy (int i, int j, int ip, int jp, int *sequence)
// returns the free energy of the internal loop closed at (i,j,ip,jp)
{

    int energy;
    int penalty_size, asym_penalty, ip_jp_energy, i_j_energy;
    int branch1, branch2, l;


    branch1 = ip-i-1;
    branch2 = j-jp-1;

    if (branch1 != 0 || branch2 != 0)
    {
        // check if it is a bulge loop of size 1
        // check if it is int11 or int21 or int22
        if (branch1 == 1 && branch2 == 1 && !simple_internal_energy)     // it is int11
        {
                        // int11[i][j][i+1][j-1][ip][jp]
            energy = int11 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]];
        }
        else if (branch1 == 1 && branch2 == 2 && !simple_internal_energy)
        {
            // int21[i][j][i+1][j-1][ip][jp][jp+1]
            energy = int21 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]

                           [sequence[ip]][sequence[jp]]
                           [sequence[jp+1]];
        }
        else if(branch1 == 2 && branch2 == 1 && !simple_internal_energy)
        {
            // after rotation: int21[jp][ip][j-1][ip-1][j][i][i+1]
            energy = int21 [sequence[jp]][sequence[ip]]
                           [sequence[j-1]][sequence[ip-1]]
                           [sequence[j]][sequence[i]]
                           [sequence[i+1]];
        }
        else if (branch1 == 2 && branch2 == 2 && !simple_internal_energy)
        {
            // int22[i][j][i+1][j-1][ip][jp][ip-1][jp+1]
            energy = int22 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]]
                           [sequence[ip-1]][sequence[jp+1]];
        }
        else
        {
            // this case is not int11, int21, int22

            // check if it is a bulge
            if (branch1 == 0 || branch2 == 0)
            {
                l = branch1+branch2;
                penalty_size = penalty_by_size (l, 'B');
                if (l == 1)
                {
                    // bulge of size 1
                    // stack[i][j][i+1][j-1]
                    energy = stack [sequence[i]][sequence[j]]
                                   [sequence[ip]][sequence[jp]] +
                               penalty_size;
                }
                else
                {
                    // bulge of size bigger than 1
                    // check if (i,j) and (ip,jp) can pair
                    energy = penalty_size +
                               AU_penalty (sequence[i],sequence[j]) +
                               AU_penalty (sequence[ip], sequence[jp]);
                }
            }
            // it is an internal loop (not a bulge)
            else
            {
                l = branch1+branch2;
                penalty_size = penalty_by_size (l, 'I');
                asym_penalty = asymmetry_penalty (branch1, branch2);

                if ((branch1 == 1 || branch2 == 1) && misc.gail_rule)
                // If gail_rule is set to 1 in miscloop file,
                // i_j_energy and ip_jp_energy will be calculated as if it was a loop of As
                {
                    i_j_energy  =  tstacki[sequence[i]][sequence[j]]
                                          [0][0];
                    ip_jp_energy = tstacki[sequence[jp]][sequence[ip]]
                                          [0][0];
                }
                else
                {
                    i_j_energy   = tstacki[sequence[i]][sequence[j]]
                                          [sequence[i+1]][sequence[j-1]];
                    ip_jp_energy = tstacki[sequence[jp]][sequence[ip]]
                                          [sequence[jp+1]][sequence[ip-1]];
                }
                energy = i_j_energy + ip_jp_energy + penalty_size +
                                         asym_penalty;
            }
        }
    }
    return energy;
}




int s_internal_loop::get_enthalpy (int i, int j, int ip, int jp, int *sequence)
// returns the enthalpy of the internal loop closed at (i,j,ip,jp)
{
    int energy;
    int penalty_size, asym_penalty, ip_jp_energy, i_j_energy;
    int branch1, branch2, l;

    branch1 = ip-i-1;
    branch2 = j-jp-1;

    if (branch1 != 0 || branch2 != 0)
    {
        // check if it is a bulge loop of size 1
        // check if it is int11 or int21 or int22
        if (branch1 == 1 && branch2 == 1 && !simple_internal_energy)     // it is int11
        {
                        // int11[i][j][i+1][j-1][ip][jp]
            energy = enthalpy_int11 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]];
        }
        else if (branch1 == 1 && branch2 == 2 && !simple_internal_energy)
        {
            // int21[i][j][i+1][j-1][ip][jp][jp+1]
            energy = enthalpy_int21 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]]
                           [sequence[jp+1]];
        }
        else if(branch1 == 2 && branch2 == 1 && !simple_internal_energy)
        {
            // after rotation: int21[jp][ip][j-1][ip-1][j][i][i+1]
            energy = enthalpy_int21 [sequence[jp]][sequence[ip]]
                           [sequence[j-1]][sequence[ip-1]]
                           [sequence[j]][sequence[i]]
                           [sequence[i+1]];
        }
        else if (branch1 == 2 && branch2 == 2 && !simple_internal_energy)
        {
            // int22[i][j][i+1][j-1][ip][jp][ip-1][jp+1]
            energy = enthalpy_int22 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]]
                           [sequence[ip-1]][sequence[jp+1]];
        }
        else
        {
            // this case is not int11, int21, int22

            // check if it is a bulge
            if (branch1 == 0 || branch2 == 0)
            {
                l = branch1+branch2;
                penalty_size = penalty_by_size_enthalpy (l, 'B');
                if (l == 1)
                {
                    // bulge of size 1
                    // stack[i][j][i+1][j-1]
                    energy = enthalpy_stack [sequence[i]][sequence[j]]
                                   [sequence[ip]][sequence[jp]] +
                               penalty_size;
                }
                else
                {
                    // bulge of size bigger than 1
                    // check if (i,j) and (ip,jp) can pair
                    energy = penalty_size +
                               AU_penalty_enthalpy (sequence[i],sequence[j]) +
                               AU_penalty_enthalpy (sequence[ip], sequence[jp]);
                }
            }
            // it is an internal loop (not a bulge)
            else
            {
                l = branch1+branch2;
                penalty_size = penalty_by_size_enthalpy (l, 'I');
                asym_penalty = asymmetry_penalty_enthalpy (branch1, branch2);

                if ((branch1 == 1 || branch2 == 1) && enthalpy_misc.gail_rule)
                // If gail_rule is set to 1 in miscloop file,
                // i_j_energy and ip_jp_energy will be calculated as if it was a loop of As
                {
                    i_j_energy  =  enthalpy_tstacki[sequence[i]][sequence[j]]
                                          [0][0];
                    ip_jp_energy = enthalpy_tstacki[sequence[jp]][sequence[ip]]
                                          [0][0];
                }
                else
                {
                    i_j_energy   = enthalpy_tstacki[sequence[i]][sequence[j]]
                                          [sequence[i+1]][sequence[j-1]];
                    ip_jp_energy = enthalpy_tstacki[sequence[jp]][sequence[ip]]
                                          [sequence[jp+1]][sequence[ip-1]];
                }
                energy = i_j_energy + ip_jp_energy + penalty_size +
                                         asym_penalty;
            }
        }
    }
    return energy;
}


void s_internal_loop::count_get_energy (int i, int j, int ip, int jp, int *sequence, int *counter)
// this function is needed for parameter learning, not for folding
// fill the counter vectro accordingly
// Mirela: Nov 23, 2003
{
    int energy;
    int penalty_size, asym_penalty, ip_jp_energy, i_j_energy;
    int branch1, branch2, l;
    char type[100];
    int index;
    branch1 = ip-i-1;
    branch2 = j-jp-1;
             
    if (branch1 != 0 || branch2 != 0)
    {
        // check if it is a bulge loop of size 1
        // check if it is int11 or int21 or int22
        if (branch1 == 1 && branch2 == 1 && !simple_internal_energy)     // it is int11
        {
                        // int11[i][j][i+1][j-1][ip][jp]
            energy = IGINF(int11 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]]);
            if (sequence[i]*100000 + sequence[j]*10000 + sequence[i+1]*1000 + sequence[j-1]*100 + sequence[ip]*10 + sequence[jp] >
                sequence[jp]*100000 + sequence[ip]*10000 + sequence[j-1]*1000+ sequence[i+1]*100 + sequence[j]*10 + sequence[i])
              {
                sprintf (type, "int11[%d][%d][%d][%d][%d][%d]",
                         sequence[jp], sequence[ip], sequence[j-1], sequence[i+1], sequence[j], sequence[i]);
              }
            else
              {
              sprintf (type, "int11[%d][%d][%d][%d][%d][%d]",
                       sequence[i], sequence[j], sequence[i+1], sequence[j-1], sequence[ip], sequence[jp]);
              }
            index = structure_type_index (type);
            counter[index]++;
        }
        else if (branch1 == 1 && branch2 == 2 && !simple_internal_energy)
        {
            // int21[i][j][i+1][j-1][ip][jp][jp+1]
            energy = IGINF(int21 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]]
                           [sequence[jp+1]]);
            sprintf (type, "int21[%d][%d][%d][%d][%d][%d][%d]",
                     sequence[i], sequence[j], sequence[i+1], sequence[j-1], sequence[ip], sequence[jp], sequence[jp+1]);
            index = structure_type_index (type);
            counter[index]++;
        }
        else if(branch1 == 2 && branch2 == 1 && !simple_internal_energy)
        {
            // after rotation: int21[jp][ip][j-1][ip-1][j][i][i+1]
            energy = IGINF(int21 [sequence[jp]][sequence[ip]]
                           [sequence[j-1]][sequence[ip-1]]
                           [sequence[j]][sequence[i]]
                           [sequence[i+1]]);
            sprintf (type, "int21[%d][%d][%d][%d][%d][%d][%d]",
                     sequence[jp], sequence[ip], sequence[j-1], sequence[ip-1], sequence[j], sequence[i], sequence[i+1]);
            index = structure_type_index (type);
            counter[index]++;
        }
        else if (branch1 == 2 && branch2 == 2 && !simple_internal_energy)
        {
            // int22[i][j][i+1][j-1][ip][jp][ip-1][jp+1]
            energy = IGINF(int22 [sequence[i]][sequence[j]]
                           [sequence[i+1]][sequence[j-1]]
                           [sequence[ip]][sequence[jp]]
                           [sequence[ip-1]][sequence[jp+1]]);
            if (sequence[i]*10000000 + sequence[j]*1000000 + sequence[i+1]*100000 + sequence[j-1]*10000 +
                sequence[ip]*1000 + sequence[jp]*100 + sequence[ip-1]*10 + sequence[jp+1] >
                sequence[jp]*10000000 + sequence[ip]*1000000 + sequence[jp+1]*100000 + sequence[ip-1]*10000 +
                sequence[j]*1000 + sequence[i]*100 + sequence[j-1]*10 + sequence[i+1])
              {
                sprintf (type, "int22[%d][%d][%d][%d][%d][%d][%d][%d]",
                         sequence[jp], sequence[ip], sequence[jp+1], sequence[ip-1],
                         sequence[j], sequence[i], sequence[j-1], sequence[i+1]);
              }
            else
              {
                sprintf (type, "int22[%d][%d][%d][%d][%d][%d][%d][%d]",
                         sequence[i], sequence[j], sequence[i+1], sequence[j-1],
                         sequence[ip], sequence[jp], sequence[ip-1], sequence[jp+1]);
              }
            index = structure_type_index (type);
            counter[index]++;
        }
        else
        {
            // this case is not int11, int21, int22
                                                                                                                                                                                                                                                                                                                                                                                                                                    
            // check if it is a bulge
            if (branch1 == 0 || branch2 == 0)
            {
                l = branch1+branch2;
                penalty_size = penalty_by_size (l, 'B');
                count_penalty_by_size (l, 'B', counter);
                if (l == 1)
                {
                    // bulge of size 1
                    // stack[i][j][i+1][j-1]
                    energy = IGINF(stack [sequence[i]][sequence[j]]
                                   [sequence[ip]][sequence[jp]]) +
                               penalty_size;
                    if (1000*sequence[i] + 100*sequence[j] + 10*sequence[ip] + sequence[jp] >
                        1000*sequence[jp] + 100*sequence[ip] + 10*sequence[j] + sequence[i])
                      sprintf (type, "stack[%d][%d][%d][%d]", sequence[jp], sequence[ip], sequence[j], sequence[i]);
                    else
                      sprintf (type, "stack[%d][%d][%d][%d]", sequence[i], sequence[j], sequence[ip], sequence[jp]);
                    index = structure_type_index (type);
                    counter[index]++;
                }
                else
                {
                    // bulge of size bigger than 1
                    // check if (i,j) and (ip,jp) can pair
                    energy = penalty_size +
                               AU_penalty (sequence[i],sequence[j]) +
                               AU_penalty (sequence[ip], sequence[jp]);
                    count_AU_penalty (sequence[i],sequence[j], counter);
                    count_AU_penalty (sequence[ip], sequence[jp], counter);
                }
            }
            // it is an internal loop (not a bulge)
            else
            {
                l = branch1+branch2;
                penalty_size = penalty_by_size (l, 'I');
                count_penalty_by_size (l, 'I', counter);
                // TODO: to add asym penalty
                asym_penalty = asymmetry_penalty (branch1, branch2);
                count_asymmetry_penalty (branch1, branch2, counter);
                                                                                                                                                                                                                                                                                                                                                                                                                                    
                if ((branch1 == 1 || branch2 == 1) && misc.gail_rule)
                // If gail_rule is set to 1 in miscloop file,
                // i_j_energy and ip_jp_energy will be calculated as if it was a loop of As
                {
                    i_j_energy  =  IGINF(tstacki[sequence[i]][sequence[j]]
                                          [0][0]);
                    sprintf (type, "tstacki[%d][%d][0][0]", sequence[i], sequence[j]);
                    index = structure_type_index (type);
                    counter[index]++;
                    ip_jp_energy = IGINF(tstacki[sequence[jp]][sequence[ip]]
                                          [0][0]);
                    sprintf (type, "tstacki[%d][%d][0][0]", sequence[jp], sequence[ip]);
                    index = structure_type_index (type);
                    counter[index]++;
                }
                else
                {
                    i_j_energy   = IGINF(tstacki[sequence[i]][sequence[j]]
                                          [sequence[i+1]][sequence[j-1]]);
                    sprintf (type, "tstacki[%d][%d][%d][%d]", sequence[i], sequence[j], sequence[i+1], sequence[j-1]);
                    index = structure_type_index (type);
                    counter[index]++;
                    ip_jp_energy = IGINF(tstacki[sequence[jp]][sequence[ip]]
                                          [sequence[jp+1]][sequence[ip-1]]);
                    sprintf (type, "tstacki[%d][%d][%d][%d]", sequence[jp], sequence[ip], sequence[jp+1], sequence[ip-1]);
                    index = structure_type_index (type);
                    counter[index]++;
                }
                energy = i_j_energy + ip_jp_energy + penalty_size +
                                         asym_penalty;
            }
        }
    }
    //    return energy;
}


